/* FLS Mamdani for an automated car pilot
 *  GBR / 22.02.2018
 */

# include <FuzzyGbr.h> // Library to compute membership degree for 
                       //triangular, trapezoidal and gaussian membership function
FuzzyGbr MDComp;  // initialize an instance of FuzzyGbr library
                  //as MDComp - membership degree computing

// -------------------------
//In general, const is preferred over #define for defining constants.
const int NUM_MF_VISIBILITY = 3;  // number of fuzzy sets for Visibility, input 
const int NUM_MF_ROAD_QUALITY = 3; // number of fuzzy sets Road Quality, input
const int NUM_MF_SPEED = 5;    // number of fuzzy sets for Speed, output 
const int MIN_SPEED = 0; // lower bound for Speed
const int MAX_SPEED = 160;  // uper bound for Speed
// ---------------------------
const int NUM_SAMPLES = 1000;//number of samples to discretize Speed
float stepSpeed; // necessary to descretize the Speed variable
float sSpeed; //sample of the Speed; 
// -------------------------------
float accMD, accMdSample, accMdOut, speedStar; // Necessary for inference, aggregation and defuzzification
float mdOut, mdOutStarMax, mdOutStar[4]; // Necessary for inference, aggregation and defuzzification

float visibilityIn; // current value for Visibility
float roadQualityIn; // current value for RoadQuality

// -------------   Define the Knowledge Base of the fuzzy system -----------------
// ......... Define the Data base - fuzzy sets ............
// define mf for Visibility
String  visibilityMfType[NUM_MF_VISIBILITY] = {"trimf","trimf","trimf"}; // mf type
float visibilityMF[NUM_MF_VISIBILITY][3]= { {0, 0, 50},   // parameters for the 1st mf: Poor
                                          {0, 50, 100},   // 2nd mf:                    Good
                                          {50, 100, 100} }; // 3rd mf:              VeryGood
// define mf for RoadQuality
String roadQualityMfType[] ={"trapmf","trapmf","trapmf"}; // mf type                                     
float roadQualityMF[][4] = { {0, 0, 10, 45},        // parameters of 1st mf: Bad
                             {10, 45, 55, 90},      // 2nd mf:               Good
                             {55, 90, 100, 100} };   // 3rd mf:           VeryGood 
// define mf for Speed
String speedMfType[] ={"gaussmf","gaussmf","gaussmf","gaussmf","gaussmf"}; // mf type                                     
float speedMF[][2]={  {12, 20},    // parameters of 1st mf: VeryLow
                      {12, 45},    // 2nd mf:               Low                    
                      {12, 70},    // 3rd mf:               Medium
                      {12, 100},   // 4th mf:               High
                      {12, 130} }; // 5th mf:               VeryHigh
// ......... End of Data base - fuzzy sets ............

// ****** Define the Rule base ********
/* In fact define only an ordered array (rule1, rule2,...,)containing the index 
  of the output fuzzy set for each rule
  For example: 0 for the first output mf; 4 for the fifth output mf.
  SLF Rule base:
  1. If Visibility is Poor     and RoadQuality is Bad        then Speed is VeryLow   0
  2. If Visibility is Poor     and RoadQuality is Good       then Speed is Low       1
  3. If Visibility is Poor     and RoadQuality is VeryGood   then Speed is Medium    2
  4. If Visibility is Good     and RoadQuality is Bad        then Speed is Low       1
  5. If Visibility is Good     and RoadQuality is Good       then Speed is Medium    2
  6. If Visibility is Good     and RoadQuality is VeryGood   then Speed is High      3
  7. If Visibility is VeryGood and RoadQuality is Bad        then Speed is Medium    2
  8. If Visibility is VeryGood and RoadQuality is Good       then Speed is High      3
  9. If Visibility is VeryGood and RoadQuality is VeryGood   then Speed is VeryHigh  4
 */ 
int rules[] = {0, 1, 2, 1, 2, 3, 2, 3, 4};  
// ****** End of Rule Base *******

// -------------   End of the Knowledge Base of the fuzzy system -----------------
                      
float mdVisibility[NUM_MF_VISIBILITY]; // membership degree of the visibilityIn to each of its fuzzy sets
float mdRoadQuality[NUM_MF_ROAD_QUALITY]; // membership degree of the qualityRoadIn to each of its fuzzy sets
float fdRule [NUM_MF_VISIBILITY*NUM_MF_ROAD_QUALITY]; // array for firing degree of the rules 
float fdActiveRule[4]; // only firing degree of active rules (fd  > 0), maximum 4 active rules 
int   idxActiveMf[4]; // index of the mf corsponding to active rules only; corellated with fdActiveRule

int i,j,k,m;  // some indexes

// define a function that computes the membership degree (MD)
float computeMD (float inpVal, float params[], String mfType ){
float md;
if (mfType == "trimf")  {    //call triMf function in the FuzzyGbr library
    md = MDComp.triMf(inpVal, params); 
  } 
  if (mfType == "trapmf")  {  //call trapMf function in FuzzyGbr library
    md = MDComp.trapMf(inpVal, params); 
  }
  if (mfType == "gaussmf")  { //call trapMf function in the FuzzyGbr library
    md = MDComp.gaussMf(inpVal, params); 
  }  
  return md;
}

void setup() {
  // put your setup code here, to run once:
Serial.begin(9600); // start the serial port
}

void loop() {
// read and display the current inputs
Serial.println("Input Visibility* in [0; 100]");
while (Serial.available() == 0);
float visibilityIn = Serial.parseFloat(); 

Serial.println("Input RoadQuality* in [0; 100]");
while (Serial.available() == 0);
float roadQualityIn = Serial.parseFloat(); 

Serial.println(" ");
Serial.print("Visibility* = "); Serial.println(visibilityIn);
Serial.print("RoadQuality* = "); Serial.println(roadQualityIn);
Serial.println(" ");

/* Compute membership degrees for curent inputs for all its fuzzy sets */
// Visibility input
Serial.println("Membership degrees for Visibility input");
for (i = 0; i < NUM_MF_VISIBILITY; i++)  {
  mdVisibility[i] = computeMD(visibilityIn, visibilityMF[i], visibilityMfType[i]);      // call the function to compute md
Serial.print(mdVisibility[i]);  Serial.print("    ");
}
Serial.println("    "); Serial.println("    ");

// RoadQuality input
Serial.println("Membership degrees for RoadQuality input");
for (i = 0; i < NUM_MF_ROAD_QUALITY; i++)  {
   mdRoadQuality[i] = computeMD(roadQualityIn, roadQualityMF[i], roadQualityMfType[i]); // call the function to compute md
Serial.print(mdRoadQuality[i]);  Serial.print("    ");
}
Serial.println("    "); Serial.println("    ");

// compute the firing degree of the rules
k = 0; m = 0;
Serial.println("Firing degree of the rules");
for (i = 0; i < NUM_MF_VISIBILITY; i++) {
  for (j = 0; j < NUM_MF_ROAD_QUALITY; j++) {
    fdRule[k] = mdVisibility[i] * mdRoadQuality[j];               // prod for AND operator for multiple antecedents
    Serial.print(fdRule[k]);  Serial.print(" ");
    if (fdRule[k]){
      fdActiveRule[m] = fdRule[k];
      idxActiveMf[m] = rules[k];
      m++;
    }
    k++;
  }
}
Serial.println("    "); Serial.println("    ");
Serial.println("Firing degree of active rules only:");
for (i = 0; i<4; i++){
Serial.print(fdActiveRule[i]); Serial.print("  ");
}
Serial.println("    "); Serial.println("    ");
Serial.println("Index of output mf for active rules only:");
for (i = 0; i<4; i++){
Serial.print(idxActiveMf[i]); Serial.print("  ");
}
// ~~~~~~~~~~~~~~~~ Start of Inference, Agreggation, Defuzzification ~~~~~~~~~~~~~
 stepSpeed = (MAX_SPEED - MIN_SPEED)/(float) NUM_SAMPLES;//discretise the Speed  
 accMD = 0;  
 accMdSample = 0;
 for (sSpeed = MIN_SPEED; sSpeed <= MAX_SPEED; sSpeed = sSpeed + stepSpeed){
   for (i = 0; i < 4; i++) {
   mdOut=computeMD(sSpeed,speedMF[idxActiveMf[i]],speedMfType[idxActiveMf[i]]);
   mdOutStar[i] = min(fdActiveRule[i], mdOut); 
    }
  mdOutStarMax = 0;
  for (i = 0; i < 4; i++){
    if (mdOutStar[i] > mdOutStarMax){
      mdOutStarMax = mdOutStar[i];
    }
  }
  accMD = accMD + mdOutStarMax;
  accMdSample = accMdSample + mdOutStarMax * sSpeed;
 }
 speedStar = accMdSample / accMD;
 Serial.println("  "); Serial.println("  ");
 Serial.print("The current Speed* is:    "); Serial.println(speedStar);
// ~~~~~~~~~~~~~~~~ End of Inference, Agreggation, Defuzzification ~~~~~~~~~~~~~~~~~~~~

Serial.println("    ");
Serial.println("===========================================================================");
Serial.println("    "); Serial.println("    ");
exit(0); // exit the loop
}

